/* 
 *  $Id: Application.Java,v 1.3 1999/09/07 21:15:53 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTS_ANTS_PKG;

import JANOS_ANTS_UTILS_PKG.KeyArgs;
import JANOS_ANTS_UTILS_PKG.InvalidKeyArgsException;

/**
 * Abstract node application.  An Application is an end-point for capsules.
 * Capsules are sent and received via the Node associated with the Application.
 * <p>
 * Applications understand the '-port' KeyArg for setting their port address
 * on the Node.
 * <p>
 * Applications are started after the Node has completed initialization.
 *
 * @author David Wetherall
 */
public abstract class Application
	extends Entity
{
	/**
	 * Default KeyArgs for an application.  Sets port to 0.
	 */
	final public static String[] defaults = { "-port", "0" };

	/**
	 * The Node associated with this application.  All Capsules
	 * are sent via and recieved from this Node.
	 *
	 * @see #setNode()
	 * @see #getNode()
	 */
	private Node node;

	/**
	 * The default default resource limit for packets
	 * sent by an Application.
	 *
	 * @deprecated Please use DEFAULT_TTL instead.
	 */
	private final static int DEFAULTRLIMIT = 64;

	/**
	 * The default TTL for packets sent by this application.
	 */
	private final static int DEFAULT_TTL = 64;
	
	
	/**
	 * The default resource limit for Capsules sent by
	 * this application.
	 *
	 * @see #setDefaultResources(int)
	 * @see #getDefaultResources()
	 */
	private int ttl;

	public int getDefaultTTL()
	{
		return ttl;
	}
	
	public void setDefaultTTL(int ttl)
	{
		this.ttl = ttl;
	}

	/**
	 * @deprecated use getDefaultTTL()
	 */
	public int getDefaultResources()
	{
		return getDefaultTTL();
	}

	/**
	 * @deprecated use setDefaultTTL(int)
	 */
	public void setDefaultResources(int r)
	{
		setDefaultTTL(r);
	}

	public int getPort()
	{
		return port;
	}

	/**
	 * Return the Node interface customized for this application.
	 */
	public Node thisNode() {
		return node; 
	}

	/**
	 * Assign the node interface to this application.  Done
	 * only in AppBooster.run().
	 *
	 * @see AppBooster#run()
	 */
	PACKAGE void setNode(Node node)
	{
		this.node = node;
	}
  

	/**
	 * The port for this application.  This is the identifier by
	 * which remote nodes can send capsules to this application:
	 * the address determines the Node, and the port determines
	 * the application.
	 * <p>
	 * Set by the '-port' KeyArg.  Defaults to 0.
	 * 
	 * @see Node#attachApplication()
	 */
	protected short port = 0;

	/**
	 * Send the given capsule with the given resource limit.
	 * <p>
	 * The capsule is primed (that is, the 'from address' and ttl
	 * are set), and then sent via the associated Node.
	 *
	 * @param cap the Capsule to send.  XXX state of Xdr?
	 * @param ttl the TTL to give the Capsule.
	 * @see Node#send(Capsule)
	 * @see #prime(int,int)
	 */
	public boolean send(Capsule cap, int ttl) {
		cap.prime(node.getAddress(), ttl); // privileged prime()
		return node.send(cap);
	}
	
	/**
	 * Send the given capsule.
	 * <p>
	 * The capsule is primed with the default TTL (this.ttl)
	 * then, the capsule is sent via the associated Node.
	 *
	 * @see #ttl
	 * @see #prime(int,int)
	 * @see Node#send(Capsule)
	 */
	public boolean send(Capsule cap) {
		cap.prime(node.getAddress(), this.ttl); // privileged prime()
		return node.send(cap);
	}
  
	/**
	 * Callback for packets that are delivered to this application.
	 * This does nothing, so the default behavior for the application
	 * is to drop delivered packets on the floor.
	 *
	 * @param cap the Capsule that got delivered.  XXX state of Xdr?
	 */
	public void receive(Capsule cap) {
		// Log("Application.receive() got " + cap);
	}
  
	/**
	 * Start this Application rolling.  This is invoked once by the
	 * Node after it has been initialized.
	 * <p>
	 * Note that implementations of start should just set things up.  If
	 * the application wants to do things like send packets, this
	 * function should start a new thread to do that.  Creating a GUI
	 * and associating callbacks is a good thing to do here.
	 * <p>
	 * XXX mostly stuff that could just be done in the no-arg 
	 * constructor, no?
	 */
	public abstract void start()
		throws Exception;
	
	/**
	 * Parse command line arguments for this application.  Understands
	 * the -port option.
	 *
	 * @see #port
	 * @see ManagedObject#setArgs(utils.KeyArgs)
	 */
	public void setArgs(KeyArgs k) 
		throws InvalidKeyArgsException
	{
		k.merge(defaults);

		for (int i = 0; i < k.length(); i++) {
			if (k.key(i).equals("-port")) {
				port = (short)Integer.parseInt(k.arg(i));
				k.strike(i);
			}
		}
    
		KeyArgs.noMore(k);
	}
  
	/**
	 * All applications must have a no-arg constructor as the
	 * Node invokes Class.newInstance() to create the application.
	 */
	public Application() {
		ttl = DEFAULT_TTL;
	} 
}
