/* 
 *  $Id: DLRequestCapsule.Java,v 1.5 1999/09/07 21:15:54 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTS_ANTS_PKG;

import java.util.Vector;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Demand Load Request capsule.
 * <p>
 * DLRequestCapsules are created in only one place:
 * <code>DLBootstrapCapsule.evaluate()</code>.  See DLBootstrapCapsule
 * for an explaination of why it will create DLRequestCapsules.
 * <p>
 * The DLRequestCapsule adds two fields to the base Capsule:
 * <ul>
 *  <li> the typeid of the desired CapsuleClass (this.wantedID)
 *  <li> the mask of chunks in the CodeGroup that are also needed (int)
 * </ul>
 *
 * @see CodeGroup
 * @see DLProtocol
 * @see DLBootstrapCapsule
 * @see DLResponseCapsule
 * @author David Wetherall
 */
final class DLRequestCapsule
  extends Capsule 
{

	// WARNING: this is a builtin capsule, whose version number
	// must be manually kept in synch with  its functionality
	public final static String VERSION = "JANOS DLRequestCapsule July 17, 1999";

	PACKAGE final static byte[] MID = findMID(JANOS_ANTS_ANTS_PKG_S + ".DLRequestCapsule");
	private final static byte[] PID = findPID(JANOS_ANTS_ANTS_PKG_S + ".DLRequestCapsule");

	protected byte[] mid()
	{
		return MID;
	} 

	protected byte[] pid()
	{
		return PID;
	}

	/**
	 * The TypeID of the CapsuleClass this Capsule is intended to
	 * load.  This is set only by the constructor (for outgoing
	 * DLRequestCapsules) or by the decode() function (for
	 * incoming DLRequestCapsules).
	 */
	private TypeID wantedID;

	/**
	 * The mask of chunks the requesting node needs to finish the
	 * appropriate CodeGroup.
	 */
	private int wantedChunkMask;
  
	public int length() {
		return super.length() + Xdr.TYPEID + Xdr.INT;
	}
  
	public Xdr encode() {
		Xdr xdr = super.encode();
		xdr.PUT(wantedID);
		xdr.PUT(wantedChunkMask);

		return xdr;
	}

	public Xdr decode() {
		Xdr xdr = super.decode();
		wantedID = xdr.TYPEID();
		wantedChunkMask = xdr.INT();

		return xdr;
	}


	/**
	 * Evaluate a DLRequestCapsule.  Based on the wantedID stored
	 * in this capsule, lookup the CapsuleClass on the local node.
	 * If the CapsuleClass is not available drop the request on
	 * the floor.  Otherwise, send DLResponseCapsules for each and
	 * every chunk the requestor doesn't have the in appropriate
	 * <emph>CodeGroup</emph>.
	 * <p>
	 * Note that CapsuleClasses for the builtin CodeGroups are not
	 * transferrable.
	 *
	 * @see CodeGroup
	 */
	public boolean evaluate(Node node) {
		boolean ret = true;

		// Look the desired method up on the local node
		CapsuleClass cc = node.lookupCapsuleClass(wantedID);

		if ((cc == null) || (cc.cl == null))
		{
			/*
			 * XXX I assume cc==null || cc.cl == null means that
			 * the capsule isn't yet loaded on this node, so
			 * we can't handle the request...
			 */
			TRACE(Trace.ANTS_DL, "DLRequest: Node " +NodeAddress.toString(getSrc())+ " wanted capsule that's not ready to go.  Dropping request.");

			return true;
		}

		TRACE(Trace.ANTS_DL, "DLRequest for " +wantedID+ " from " +NodeAddress.toString(getSrc())+ " for chunks 0x" +Integer.toHexString(wantedChunkMask));
		
		CodeGroup g = cc.group;
		
		// test to see if this is builtin and non-transferable
		if (g.builtin()) {
			TRACE(Trace.ANTS_DL, "DLRequest for builtin class " +cc.name()+ " discarded");

			// discard
			return true;
		}
		
		/*
		 * Okay, send off one DLResponseCapsule for each chunk that
		 * the remote node requested.
		 * 
		 * XXX Instead of looping through all the chunks to see if
		 * they want it, probably should loop through the desired
		 * chunk mask and send the appropriate chunk off directly.
		 * Not a big deal, most CodeGroups don't have a lot of chunks.
		 */
		int idx = 0;  // index of the chunk we're checking out

		cc = null; // don't get confused by eariler cc
		
		for (int i = 0; i < g.classes.size(); i++) 
		{
			cc = (CapsuleClass)g.classes.elementAt(i);
			for (int p = 0;
			     p < cc.codeSize();
			     p += CodeGroup.CHUNK_SIZE) 
			{
				if ((wantedChunkMask | g.chunkMask(idx)) !=  wantedChunkMask) {
					idx++; // don't want this chunk, try next.
					continue;
				}
				
				// Figure out how big the chunk we're sending is
				int l = cc.codeSize() - p;
				if (l > CodeGroup.CHUNK_SIZE)
					l = CodeGroup.CHUNK_SIZE;
				
				TRACE(Trace.ANTS_DL, "DLRequest: sending " +p/CodeGroup.CHUNK_SIZE+ "th chunk of " +cc.name()+ " (group mask=" +Integer.toHexString(g.chunkMask(idx))+ "; size=" +l+ ")");
				
				/*
				 * Construct a DLResponseCapsule with this chunk of
				 * CapsuleClass cc and send it off.
				 */
				Capsule dl = new DLResponseCapsule(cc, idx, p, l); 
				dl.prime(this); // XXX dl.src == this.src??!  seems wrong to me...
				boolean sent = node.routeForNode(dl, getPrevious());
				ret = ret && sent;
				
				idx++; // bump chunk number
			}
		}
		
		return ret;
	}
  
	/**
	 * Create a blank DLRequestCapsule.
	 * <p>
	 * This constructor should only be invoked when
	 * a DLRequestCapsule is being received and decoded.
	 */
	DLRequestCapsule()
	{
	}

	/**
	 * Create a DLRequestCapsule for a specific MID.
	 * <p>
	 * All chunks of the relevant CodeGroup will be requested.
	 * <p>
	 * Only invoked by ProtocolBooster.startNewMid()
	 *
	 * @see ProtocolBooster#startNewMid(Node,TypeID,int)a
	 */
	DLRequestCapsule(TypeID mid)
	{
		wantedID = mid;
		wantedChunkMask = 0xffffffff;
	}

}
