/*
 * ProtocolBooster.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTS_ANTS_PKG;

import JANOS_CORE_PKG.Address;
import JANOS_CORE_PKG.BufferHandle;
import JANOS_CORE_PKG.CommQueue;
import JANOS_CORE_PKG.DemultiplexKey;
import JANOS_CORE_PKG.Flow;
import JANOS_CORE_PKG.FlowRun;
import JANOS_CORE_PKG.ThreadPool;
import JANOS_CORE_PKG.InChannel;
import JANOS_CORE_PKG.InChannelHandle;
import JANOS_CORE_PKG.CommSpaceHandle;
import JANOS_CORE_PKG.CommSpaceElement;

import java.io.IOException;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * This is invoked by PrimordialNode as the intial class in a new
 * flow.  This new flow is charged with taking the new (unknown
 * protocol) packet, requesting and loading the code associated with
 * that packet and then processing the packet.
 * <p>
 * CommSpaceHandle should point to a valid CSH for
 * the Booster class, plus:
 *   [2] a byte[].  The MID of the protocol this Booster is working for.
 *   [3] an integer -- the logical ANTS address of the node that sent the packet.
 *   [4] a slot to write an InChannelHandle into.
 *
 */
public class ProtocolBooster
	extends Booster
{
	public ProtocolBooster(JANOS_CORE_PKG.ClassInstantiationKey key)
		throws SecurityException
	{
		key.checkEquals(ProtocolBooster.class);
	}

	public void run(Flow flow,
			ThreadPool threadPool,
			String[] args,
			CommSpaceHandle commSpaceHandle)
		throws Exception
	{
		assert(flow != null);
		assert(threadPool != null);
		assert(commSpaceHandle != null);

		assert(args == null); // not expecting any

		TRACE(Trace.ANTS_DL, "Starting unknown protocol.");

		DEBUGONLY(flow.setName("Flow for Unknown Protocol"));

		final Node node = setupLocalNode(flow, commSpaceHandle);

		// Dedicate a thread to handling new MIDs:
		final CommQueue queue = commSpaceHandle.read(3).getCommQueue();
		threadPool.threadStart(new Runnable() {
				public void run() {newMidLoop(queue, node);}
			}
		);
	}

	private void newMidLoop(CommQueue queue, Node node)
	{
		for(;;)
		{
			try
			{
				byte[] midBytes =
					queue.waitToTake().getByteArray();
				int prev = queue.waitToTake().getInt();
				startNewMid(node, TypeID.wrap(midBytes), prev);
			}
			catch(Throwable e)
			{
				e.printStackTrace();
			}
		}
	}

	// It's possible for a MID to get started twice.  But this
	// is probably ok, and maybe even necessary: if the first
	// installation fails halfway through, a second installation
	// has to complete it.
	private void startNewMid(Node node, TypeID mid, int prev)
	{
		// Setup a Key for receving DLReply packets...
		new CapsuleChannel(DLResponseCapsule.class,
			new CodeRequestDemultiplexKey(
				DLResponseCapsule.MID,
				mid.unwrap()));

		// Construct a DLRequest packet and send it to prevNode.
		DLRequestCapsule req = new DLRequestCapsule(mid);
		req.prime(node.getAddress(), 1);
		node.routeForNode(req, prev);
	}
}
