/*
 * UnknownMidPackets.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTS_ANTS_PKG;

import JANOS_CORE_PKG.BufferHandle;
import JANOS_CORE_PKG.CommSpaceHandle;
import JANOS_CORE_PKG.CommSpaceElement;
import JANOS_CORE_PKG.Flow;
import JANOS_CORE_PKG.InChannelHandle;

IMPORT_DEBUG

/**
 * UnknownMidPackets is a temporary storage place for
 * packets of unknown MIDs to live in while the MID is
 * being resolved.  Although such a storage space
 * isn't essential for correctness (we could just drop
 * the packets and then have them retransmitted after
 * the MID is resolved), holding the packets should
 * smooth performance a bit.  unknownMidPackets maps
 * MIDs to arrays of BufferHandles.
 * <p>
 * In the absence of any decent synchronization mechanisms, we
 * make unknownMidPackets a lock-free data structure.
 * unknownMidPackets is a CommSpace consisting of a number
 * of smaller "spaces".	 Each space is a CommSpace with
 * the following elements:
 * <pre>
 *   0: the MID whose packets are held in this space
 *   1: the number of buffered packets in this space
 *   2, 3, ...: data items
 * </pre>
 * Each data item is either null (uninitialized),
 * a BufferHandle holding a packet, or an InChannel onto
 * which packets should be injected.  The server tries
 * to insert packets into the space, and the flow tries
 * to initialize an InChannel and fill all the data items
 * with the InChannel handle.  Whoever gets to the space
 * first simply inserts their item.  Whoever gets to the
 * space second takes the item that was already there, together
 * with their own item, so that they have both a packet and
 * an InChannel, and pushes the packet onto the InChannel.
 * An atomic swap is used to insert things into the space,
 * so that the server/flow can tell whether they got to
 * each item first or second.
 * <p>
 * This is both the client and server end (via constructor).
 */
PACKAGE class UnknownMidPackets
{
	PACKAGE final CommSpaceHandle spaces;
	private int nextSpace;
	private final static int NUMSPACES = 20;
	private final static int PACKETS_PER_SPACE = 5;

	PACKAGE final static int NEVER_DEMULTIPLEXED = -1;

	PACKAGE UnknownMidPackets()
	{
		spaces = CommSpaceHandle.newCommSpace(NUMSPACES);
		byte[] b = new byte[0];
		for(int i = 0; i < NUMSPACES; i++)
		{
			CommSpaceHandle space =
				CommSpaceHandle.newCommSpace(
					2 + PACKETS_PER_SPACE);

			space.write(0, null);
			spaces.write(i, new CommSpaceElement(space));
		}
	}

	PACKAGE UnknownMidPackets(CommSpaceHandle spaces)
	{
		this.spaces = spaces;
	}

	// This is deliberately not synchronized.  It is safe for
	// concurrent execution.
	private CommSpaceHandle spaceForMID(byte[] midBytes)
	{
		byte[] bytes = new byte[midBytes.length];

		for(int i = 0; i < NUMSPACES; i++)
		{
			CommSpaceHandle space = spaces.read(i)
				.getCommSpaceHandle();
			
			CommSpaceElement elem = space.read(0);

			if (elem == null)
				continue;

			elem.getByteArray(bytes, 0);
			int j = 0;
			for(; j < bytes.length; j++)
			{
				if(bytes[j] != midBytes[j])
					break;
			}
			if(j == bytes.length)
				return space;
		}
		
		return null;
	}

	// Returns true if the mid was already in the cache
	PACKAGE synchronized boolean addPacket(byte[] midBytes,
		BufferHandle packet)
	{
		try
		{
			// Does the MID already have a space?
			CommSpaceHandle space = spaceForMID(midBytes);

			// If the MID doesn't have a space, make one (and
			// kick out some other unlucky MID in sort of an LRU
			// fashion):
			boolean found = (space != null);
			if(!found)
			{
				// Clear out the old space:
				CommSpaceHandle old = spaces.read(nextSpace)
					.getCommSpaceHandle();

				for(int i = 0; i < PACKETS_PER_SPACE; i++)
				{
					Object o = old.swap(2 + i, null);

					// Manually clean out the unused BufferHandle
					if(o instanceof BufferHandle)
						((BufferHandle) o).reset();
				}
	
				// And put in a new space:
				space = CommSpaceHandle.newCommSpace(
					2 + PACKETS_PER_SPACE);
				space.write(0, new CommSpaceElement(midBytes));
				space.write(1, new CommSpaceElement(0));
	
				spaces.write(nextSpace, new CommSpaceElement(space));
				nextSpace = (nextSpace + 1) % NUMSPACES;
			}
	
			// Where does the next packet go in the space, and
			// is the space full?
			int nextPacket = space.read(1).getInt();
			if(nextPacket >= PACKETS_PER_SPACE)
			{
				return found;
			}

			space.write(1, new CommSpaceElement(nextPacket + 1));

			// Swap the packet in:
			BufferHandle h = new BufferHandle(Flow.currentFlow(),
				packet);
			CommSpaceElement elem = space.swap(
				2 + nextPacket,
				new CommSpaceElement(h));
			if(elem != null)
			{
				// The InChannel has already been established,
				// so go ahead and push the packet onto it.
				elem.getInChannelHandle().push(h,
					NEVER_DEMULTIPLEXED);

				// Release the packet now rather than waiting
				// for the "Clear out the old space" code
				// above to release it.
				space.write(2 + nextPacket, null);
				h.reset();
			}
			return found;
		}
		catch(Exception e)
		{
			FATALEXCEPTION(e, "Blew up adding packet to UnkonwnMidPacketBag");
		}
	}

	// This is deliberately not synchronized.  It is safe for
	// concurrent execution.
	PACKAGE void setInChannel(byte[] midBytes, InChannelHandle inChannel)
	{
		CommSpaceHandle space = spaceForMID(midBytes);
		if(space == null) return;

		CommSpaceElement inChan = new CommSpaceElement(inChannel);

		// Swap the InChannel in:
		for(int i = 0; i < PACKETS_PER_SPACE; i++)
		{
			CommSpaceElement elem = space.swap(
				2 + i, inChan);
			if(elem != null)
			{
				BufferHandle packet = elem.getBufferHandle();
				inChannel.push(packet, NEVER_DEMULTIPLEXED);
				packet.reset();
			}
		}
	}
}
