/* 
 *  $Id: KeyArgs.Java,v 1.3 1999/09/07 21:16:58 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTS_UTILS_PKG;

IMPORT_DEBUG

/**
 * a (key, argument) pair.
 *
 * @author David Wetherall
 */
PACKAGE class KeyArg {
  String key;
  String arg;

  /**
   * Construct a new (key, argument) pair.
   *
   * @param k the key
   * @param a the argument
   */
  KeyArg(String k, String a) { key = k; arg = a; }
}


/**
 * Keyword argument manipulation utilities
 *
 * @author David Wetherall
 */
public class KeyArgs {
  private KeyArg[] args;

  /**
   * The canonical 'null argument' KeyArg.
   */
  private static KeyArg empty = new KeyArg("", null);

  public int length() { return args.length; } 
  public String key(int i) { return args[i].key; }
  public String arg(int i) { return args[i].arg; }

  public String get(String key) {
    for (int i = 0; i < args.length; i++)
      if (args[i].key.equals(key))
	return args[i].arg;
    return null;
  }

  public void set(String key, String arg) {
    for (int i = 0; i < args.length; i++)
      if (args[i].key.equals(key)) {
	args[i].arg = arg;
	return;
      }
    KeyArg[] ret = new KeyArg[args.length + 1];  
    for (int j = 0; j < args.length; j++)
      ret[j] = args[j];

    ret[args.length] = new KeyArg(key, arg);
    args = ret;
  }
  
  public void strike(int i) { args[i] = empty; }

  public void compact() {
    int len = 0;
    for (int i = 0; i < args.length; i++)
      if (args[i] != empty) len++;

    if (len == args.length) return;

    KeyArg[] ret = new KeyArg[len];
    int idx = 0;
    for (int i = 0; i < args.length; i++)
      if (args[i] != empty)
	ret[idx++] = args[i];

    args = ret;
  }
  
  public void merge(String[] ka) 
    throws InvalidKeyArgsException
  {
    merge(new KeyArgs(ka));
  }

  public void merge(String[] ka, int skip) 
    throws InvalidKeyArgsException
  {
    merge(new KeyArgs(ka, skip));
  }

  public void merge(KeyArgs k) {
    compact();

    if (k == null)
      return;
    
    k.compact();
	
    int len = args.length;
    for (int i = 0; i < k.args.length; i++) {
      boolean match = false;
      for (int j = 0; j < args.length; j++)
	if (args[j].key.equals(k.args[i].key)) {
	  match = true;
	  break;
	}
      if (!match) len++;
    }
    
    KeyArg[] ret = new KeyArg[len];  
    for (int j = 0; j < args.length; j++)
      ret[j] = args[j];
    
    int idx = args.length;
    for (int i = 0; i < k.args.length; i++) {
      boolean match = false;
      for (int j = 0; j < args.length; j++)
	if (args[j].key.equals(k.args[i].key)) {
	  match = true;
	  break;
	}
      if (!match) ret[idx++] = k.args[i];
    }

    args = ret;
  }

  public String toString() {
    compact();
    String ka = "";
    for (int i = 0; i < args.length; i++) {
      ka += "\"" + args[i].key + "\" ";
      if (args[i].arg != null) ka += "\"" + args[i].arg + "\" ";
    }
    return ka;
  }
  
  public String[] toStrings() {
    compact();
    int len = args.length;
    for (int i = 0; i < args.length; i++)
      if (args[i].arg != null) len++;

    String[] ret = new String[len];

    int idx = 0;
    for (int i = 0; i < args.length; i++) {
      ret[idx++] = args[i].key;
      if (args[i].arg != null)
	ret[idx++] = args[i].arg;
    }

    return ret;
  }

  public static void noMore(KeyArgs k)
  {
    // XXX for debugging:

    k.compact();
    if (k.length() > 0)
      panic("leftover key args: " +k);
  }

  private KeyArgs(KeyArg[] ka) {
    args = ka;
  }
  
  public KeyArgs(String[] ka)
    throws InvalidKeyArgsException
  {
    this(ka, 0);
  }

  public KeyArgs() 
    throws InvalidKeyArgsException
  {
    this(null, 0);
  }

  public KeyArgs(String[] ka, int skip)
    throws InvalidKeyArgsException
  {
    // handle the trivial case first
    if (ka == null || ka.length == 0 || ka.length <= skip) {
      args = new KeyArg[0];
      return;
    }
    
    // walk the list once, counting and checking syntax
    int l = 0;
    boolean ok = false;
    for (int i = skip; i < ka.length; i++) {
      if (ka[i] != null && ka[i].length() > 0 && ka[i].charAt(0) == '-') {
	l++;
	ok = true;
	continue;
      }
      if (!ok) {
	String ks = "";
	for (int j = 0; j < ka.length; j++)
	  ks += ka[j] + " ";
	throw new InvalidKeyArgsException("invalid sequence: " + ks);
      }
      ok = false;
    }
    
    // now fill it in for real 
    args = new KeyArg[l];
    int idx = 0;
    for (int i = skip; i < ka.length; i++) {
      if (ka[i] != null && ka[i].length() > 0 &&  ka[i].charAt(0) == '-')
	args[idx++] = new KeyArg(ka[i], null);
      else
	args[idx-1].arg = ka[i];
    }
  }
}

/*
 * Local Variables:
 * c-basic-offset: 2
 * End:
 */
