/* 
 *  $Id: LineListTokenizer.Java,v 1.3 1999/09/07 21:16:58 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTS_UTILS_PKG;

import java.io.StreamTokenizer;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.Reader;


/**
 * File reading utility. Returns each useful line of text as an array
 * of strings, one string per "word."  Comments are ignored and quoted
 * strings are handled (the quotes are stripped off and the quoted
 * string is returned as a single word.
 * <p>
 * Max of 256 words per line (see MAXLIST).
 *
 * @author David Wetherall
 */
public class LineListTokenizer
  extends StreamTokenizer 
{
  // sorry, no more than 256 elements per line
  private final static int MAXLIST = 256;
  private final String[] list = new String[MAXLIST];

  /**
   * Parse the next non-empty line into an array of strings.
   * Result is null when EOF, and otherwise has at least one element.
   *
   * @return an array of tokens from the line.
   */
  public String[] list() 
    throws Exception 
  {
    boolean done;
    int length;

    done = false;    
    length = 0;

    while (!done && length == 0) {
      do {
	int t = nextToken();
	
	if (t == TT_EOF)
	  done = true;
	else if (t == TT_EOL)
	  break;
	// numbers not expected but handle them anyhow
	else if (t == TT_NUMBER)
	  list[length++] = Double.toString(nval);
	else
	  list[length++] = sval;
      } while (!done && length < MAXLIST);
    }
    
    if (length == MAXLIST) throw new Exception("Line too long");
    if (done || length == 0) return null;

    final String[] result = new String[length];
    System.arraycopy(list, 0, result, 0, length);
    return result;
  }


  /**
   * Create a new LineListTokenizer for the given file.
   *
   * @param file the file to use as input.
   * @exception FileNotFoundException if the given file cannot be opened for reading.
   */
  public LineListTokenizer(String file)
    throws FileNotFoundException
  {
    this(new InputStreamReader(new FileInputStream(file)));
  }

	
  /**
   * Create a new LineListTokenizer from the given read-able
   * Stream.
   *
   * @param the Reader to stream input from.
   */
  public LineListTokenizer(Reader input)
  {
    super(input);

    resetSyntax();
    whitespaceChars(0, ' ');
    wordChars('0', '9');
    wordChars('-', '-');
    wordChars('.', '.');
    wordChars(':', ':');
    wordChars('=', '=');
    wordChars('/', '/');
    wordChars('A', 'Z');
    wordChars('a', 'z');
    commentChar('#');
    quoteChar('\'');
    quoteChar('\"');    
    eolIsSignificant(true);
  }
}

/*
 * Local Variables:
 * c-basic-offset: 2
 * End:
 */
