/*
 * CommSpaceElement.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_CORE_PKG;

/**
 * Element in the CommSpace for communication.
 * This is what gets written to the CommSpace
 * <p>
 * XXX does this need to exist or can I just stick
 * Objects in the CommSpace?
 * <p>
 * XXX because the val field is final, its impossible to
 * re-use a CommSpaceElement (i.e., for a different object...)
 * Could work around this by making the val field read-only
 * while the object was lodged in the CommSpace...
 * <p>
 * XXX these are shared between processes.... i.e., they leak (~2 words).
 *
 * "Big Deal." -- Chris H.
 */
public final class CommSpaceElement
{
	private final Object val;
	private final int type;

	private static final int INT_TYPE = 1;
	private static final int CHAR_ARRAY_TYPE = 2;
	private static final int BYTE_ARRAY_TYPE = 3;
	private static final int IN_CHANNEL_TYPE = 4;
	private static final int ADDRESS_TYPE = 5;
	private static final int BUFFER_HANDLE_TYPE = 6;
	private static final int COMM_SPACE_HANDLE_TYPE = 7;
	private static final int COMM_QUEUE_TYPE = 8;
	private static final int COMM_MONITOR_TYPE = 9;

	private CommSpaceElement(Object val, int type)
	{
		// this.owner = Flow.currentFlow();

		if(val == null) throw new NullPointerException();

		this.val = val;
		this.type = type;
	}

	PACKAGE CommSpaceElement(CommSpaceElement from)
	{
		this(from.val, from.type);
	}

	public CommSpaceElement(int val)
	{
		this(new Integer(val), INT_TYPE);
	}

	public CommSpaceElement(char[] val)
	{
		this(val, CHAR_ARRAY_TYPE);
	}

	public CommSpaceElement(byte[] val)
	{
		this(val, BYTE_ARRAY_TYPE);
	}

	public CommSpaceElement(InChannelHandle val)
	{
		this(val, IN_CHANNEL_TYPE);
	}

	public CommSpaceElement(Address val)
	{
		this(val, ADDRESS_TYPE);
	}

	public CommSpaceElement(BufferHandle val)
	{
		this(val, BUFFER_HANDLE_TYPE);
	}

	public CommSpaceElement(CommSpaceHandle val)
	{
		this(val, COMM_SPACE_HANDLE_TYPE);
	}

	public CommSpaceElement(CommQueue val)
	{
		this(val, COMM_QUEUE_TYPE);
	}

	public CommSpaceElement(CommMonitor val)
	{
		this(val, COMM_MONITOR_TYPE);
	}

	public int type()
	{
		return type;
	}

	public int getInt()
	{
		return ((Integer) val).intValue();
	}

	public char[] getCharArray()
	{
		char[] from = (char[]) val;
		char[] to = new char[from.length];
		System.arraycopy(from, 0, to, 0, from.length);
		return to;
	}

	public void getCharArray(char[] outbuf, int index)
	{
		char[] from = (char[]) val;
		System.arraycopy(from, 0, outbuf, index, from.length);
	}

	public byte[] getByteArray()
	{
		byte[] from = (byte[]) val;
		byte[] to = new byte[from.length];
		System.arraycopy(from, 0, to, 0, from.length);
		return to;
	}

	public void getByteArray(byte[] outbuf, int index)
	{
		byte[] from = (byte[]) val;
		System.arraycopy(from, 0, outbuf, index, from.length);
	}

	public InChannelHandle getInChannelHandle()
	{
		InChannelHandle ich = new InChannelHandle();
		ich.copy((InChannelHandle) val);
		return ich;
	}

	public Address getAddress()
	{
		return new Address((Address) val);
	}

	public BufferHandle getBufferHandle()
	{
		return new BufferHandle(Flow.currentFlow(),
			(BufferHandle) val);
	}

	public CommSpaceHandle getCommSpaceHandle()
	{
		return new CommSpaceHandle(
			(CommSpaceHandle) val);
	}

	public CommQueue getCommQueue()
	{
		return (CommQueue) val;
	}

	public CommMonitor getCommMonitor()
	{
		return (CommMonitor) val;
	}
}
