/*
 * DemultiplexKey.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_CORE_PKG;

/**
 * A key for matching incoming packets.
 */
public abstract class DemultiplexKey
{
	/**
	 * "length" of this filter for sorting the
	 * filters (longest first..)
	 */
	private final int length;

	public DemultiplexKey(int length)
	{
		this.length = length;
	}

	PACKAGE final int length()
	{
		return length;
	}

	/**
	 * See if the offset'th byte of the packet is val.
	 * The packet is defined as the sub-array of
	 * packetBuf from packetOffset to packetOffset + packetLength.
	 *
	 * @return 0 if it failed to match; offset+1 if it
	 * did match.
	 */
	protected static final int compareByte(byte[] packetBuf, 
					       int packetOffset,
					       int packetLength,
					       int offset,
					       byte val)
	{
		if (packetLength < offset + 1)
			return 0;

		assert(packetOffset + packetLength <= packetBuf.length);

		if (packetBuf[packetOffset + offset] == val)
			return offset + 1;

		// failed to match.
		return 0;
	}

	/**
	 * Comparse the slice of bytes at offset'th byte of the packet
	 * against the given slice.  The packet is defined as the
	 * sub-array of packetBuf from packetOffset to packetOffset +
	 * packetLength.
	 *
	 * @return offset + slice.length if they match, otherwise returns zero.
	 */
	protected static final int compareSlice(byte[] packetBuf,
						int packetOffset,
						int packetLength,
						int offset,
						byte[] slice)
	{
		if (packetLength < offset + slice.length)
			return 0;

		assert(packetOffset + packetLength <= packetBuf.length);

		int totalOffset = offset + packetOffset;

		for(int i = 0; i < slice.length; i++)
		{
			if(slice[i] != packetBuf[totalOffset + i])
				return 0;
		}

		return offset + slice.length;
	}

	/**
	 * XXX this is not supposed to be public.  It is currently
	 * public so that it may be overridden to supply more
	 * interesting packet demultiplexing matchers (e.g., in ANTS).
	 *
	 * Check if the given packet matches this filter.  Return 0
	 * if it does not, or a non-zero integer value to pass to the
	 * receive function.
	 */
	public abstract int matches(byte[] data, int packetOffset, int packetLength);


#ifdef DEBUG
	public String toString()
	{
		return this.getClass().getName()+ "[len=" +length+ "]";
	}
#endif	
}
