/*
 * InChannel.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_CORE_PKG;

import java.util.Vector;

import java.io.IOException;
import java.net.DatagramSocket;
import java.net.DatagramPacket;

#ifdef BENCH
import JANOS_LIB_PKG.CycleTimer;
import JANOS_LIB_PKG.Semaphore;
#endif

/**
 * Java wrapper for an InChannel.
 * <p>
 * Implemented on top of UDP for the moment.
 */
public final class InChannel
{
	/**
	 * Death Packet. XXX.
	 */
	public static final byte[] exitPacket = {-0x22, -0x53, -0x42, -0x11};

	/**
	 * This is a kernel-global variable.  It maps Demultiplex
	 * keys to Channels.
	 */
	private static final Vector classifierTable = new Vector();

	/**
	 * The flow this receive channel is associated with.
	 */
	PACKAGE final Flow flow;

	/**
	 * The demultiplex key associated with this channel.
	 */
	private final DemultiplexKey classifier;

	/**
	 * The ThreadPool to send incoming packets to.
	 * ThreadPool.addWork() had best be protected kernel code
	 * (i.e., final).
	 */
	private final ThreadPool threadPool;

	/**
	 * The callback for delivery.
	 */
	private final DeliverFunc deliverFunc;

	public InChannel(Flow flow,
			 DemultiplexKey classifier,
			 ThreadPool threadPool,
			 DeliverFunc deliverFunc)
	{
		this.flow = flow;
		this.classifier = classifier;
		this.threadPool = threadPool;
		this.deliverFunc = deliverFunc;

		assert(classifier);
		// XXX doesn't sufficiently protect the kernel...
		if((flow == null)
		   || (threadPool == null)
		   || (classifier == null))
			throw new NullPointerException();

		addClassifier(this);
	}
	
	private static void addClassifier(InChannel newChan)
	{
		// Insert in sorted order...
		int i;
		for (i = 0; i < classifierTable.size(); i++)
		{
			InChannel entry = (InChannel) classifierTable.elementAt(i);
			if (entry.classifier.length() <= newChan.classifier.length())
				break;
		}
		TRACE(Trace.CORE_DISPATCH, "Adding packet classifier " +newChan.classifier);
		classifierTable.insertElementAt(newChan, i);
	}

	/**
	 * XXX
	 * @see #msgLoop(int)
	 */
	public static void msgLoop(Address localAddr)
		throws IOException
	{
		assert(localAddr);
		assert(localAddr.iNetPort() != 0);
		msgLoop(localAddr.iNetPort());
	}

	/**
	 *  This must be run before any Channel objects are created.
	 *  <p>
	 *  XXX: shouldn't really be public
         *
	 *  @allocations
         */
	public static void msgLoop(int recvPort)
		throws IOException
	{
		// Set up a "packet of death":
		DemultiplexKey exitKey = new DemultiplexKey(1024) {
			public int matches(byte[] buf, int packetOffset, int packetLength) {
				return compareSlice(buf, packetOffset, packetLength,
						    0, exitPacket);
			}
		};

		DatagramSocket socket = new DatagramSocket(recvPort);

		// We need to have at least one packet around to
		// receive data before the data gets demultiplexed:
		UdpBuffer.addBuffers(1);

		for(;;)
		{
			boolean sentIt;
			UdpBuffer b = UdpBuffer.getFree();
			assert(b != null);
			
			BENCHONLY(processTimer.end());

			b.receive(socket);

			BENCHONLY(sendToRecvTimer.end());
			BENCHONLY(recvToSendTimer.start());
			BENCHONLY(recvToRecvTimer.end());
			BENCHONLY(recvToRecvTimer.start());
			BENCHONLY(processTimer.start());

			TRACE(Trace.CORE_INCOMING, "in!");
			TRACE(Trace.CORE_INCOMING, "received: " + Debug.dbgPacketSummary(b));
			sentIt = demultiplex(b);

			if(!sentIt
			   && (exitKey.matches(b.data, 0, b.length()) != 0))
				break;
			else if(!sentIt)
			{
				TRACE(Trace.CORE_INCOMING, "Can't schedule pkt: port=" +recvPort+ "; data=" +Debug.dbgPacketSummary(b));
				b.release();
			}
		}

		TRACE(Trace.CORE_INCOMING, "recv port " +recvPort+ ": exiting");
	}

	/**
	 * I sure hope this gets better
	 * <p>
	 * XXX I broke this.  The return value is true if the
	 * packet was dispatched *and* the target had enough
	 * resources to handle the packet.  It returns false
	 * if the packet wasn't matched *or* it was matched
	 * but the target didn't have enough resources.
	 */
	PACKAGE static boolean demultiplex(UdpBuffer b)
	{
		/*
		 * Assumes entries are sorted by length (longest first).
		 */
		for(int i = 0; i < classifierTable.size(); i++)
		{
			int payloadStart;
			InChannel entry = (InChannel)
				classifierTable.elementAt(i);
			assert(entry.classifier); // constructor must verify this
			payloadStart = entry.classifier.matches(b.data,
								0,
								b.length());
			if (payloadStart != 0)
			{
				TRACE(Trace.CORE_INCOMING, "demultiplexed to: " +entry.classifier);
				return entry.scheduleDelivery(b, 0, b.length(),
							      payloadStart);
			}
		}
		
		return false;
	}

	/**
	 * Schedule the given buffer b for delivery to this InChannel.
	 * A BufferHandle wrapping the packet is set up (from packetStart
	 * to packetLength).  The payloadStart parameter is just a *hint*
	 * passed to the delivery function.  To be meaningful, it should
	 * be relative to the packetStart.
	 * <p>
	 * Called by demultiplex (above) and by InChannelHandle.push().
	 *
	 * @return true if the buffer was scheduled, false if the
	 * target had insuffient resources.
	 */
	PACKAGE boolean scheduleDelivery(Buffer b, int packetStart, int packetLength,
					 int payloadStart)
	{
		// XXX I think I broke something.. 
		if(!b.charge(this.flow.state))
		{
			// flow is out of buffer space:
			TRACE(Trace.CORE_INCOMING, "out of buffer space");
			b.release();
			return false;
		}

		PacketWorker worker = PacketWorker.getFree(flow);
		assert(worker != null); // XXX 

		worker.deliverFunc = this.deliverFunc;
		worker.bufferHandle.setBuffer(b, packetStart, packetLength);

		worker.payloadStart = payloadStart;
		BENCHONLY(threadSchedTimer.start());
		threadPool.addWork(worker.workItem);
		return true;
	}

#ifdef BENCH
	public static CycleTimer processTimer = new CycleTimer("InChannel processing");
	public static CycleTimer sendToRecvTimer = new CycleTimer("InChannel send-to-recv");
	public static CycleTimer recvToSendTimer = new CycleTimer("InChannel recv-to-send");
	public static CycleTimer recvToRecvTimer = new CycleTimer("InChannel recv-to-recv");
	public static CycleTimer threadSchedTimer = new CycleTimer("InChannel thread schedule");
	public static Semaphore cheatLock = new Semaphore();
	public static byte[] cheatBuffer = new byte[65536];
	public static int cheatLen;
#endif //BENCH
}


