/*
 * UdpBuffer.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_CORE_PKG;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;

/**
 * A UDP implementation of Buffer.
 */
PACKAGE class UdpBuffer
	extends Buffer
{
	PACKAGE static UdpBuffer freeList;
	PACKAGE static final int MAX_PACKET_LEN = 65536;
	
	private final DatagramPacket udpPacket;
	PACKAGE UdpBuffer nextFree;
	private FlowState state;
	private int len;
	
	/**
	 * Add a bunch of buffers to the free buffer queue.
	 * <p>
	 * @allocation: allocates the (UdpBuffers) to be added.
	 */
	PACKAGE static synchronized void addBuffers(int numBuffers)
	{
		for(int i = 0; i < numBuffers; i++)
			addToFreeList(new UdpBuffer());
	}
	
	/**
	 * Get the next available UdpBuffer.
	 * <p>
	 * Returns null if no buffers are available.
	 */
	PACKAGE static synchronized UdpBuffer getFree()
	{
		UdpBuffer b = freeList;
		if(b != null)
			freeList = b.nextFree;
		return b;
	}
	
	/**
	 * UdpBuffers can only be allocated by addBuffers()
	 * <p>
	 * Each UdpBuffer is MAX_PACKET_LEN bytes big.
	 *
	 * @see #addBuffers()
	 */
	private UdpBuffer()
	{
		super(new byte[MAX_PACKET_LEN]);
		
		udpPacket = new DatagramPacket(data, data.length);
	}
	
	/**
	 * Charge the given flow for this buffer.  This
	 * buffer will be associated with the flow (via
	 * this.state).
	 * <p>
	 * Returns true if the given flow can pay for this buffer.
	 */
	PACKAGE boolean charge(FlowState state)
	{
		synchronized(state.lock)
		{
			if (state.freeBuffers == 0)
				return false;

			// Associate with flow so we can credit it
			// XXX assert(this.state == null);
			this.state = state;

			// Charge the flow
			state.freeBuffers--;

			return true;
		}
	}

	/**
	 * Credit the flow associated with this buffer.
	 */
	private void credit()
	{
		assert(state);
		synchronized(state.lock)
		{
			state.freeBuffers++;
		}
	}
	
	/**
	 * Put this buffer back in the recycle list (we're
	 * done with it).
	 */
	PACKAGE void recycle()
	{
		acquire();
		addToFreeList(this);
		if(state != null)
			credit();
	}

	/**
	 * Add the given flow to this global free list.
	 */
	private static synchronized void addToFreeList(UdpBuffer b)
	{
		// XXX assert(!b.onFreeList());
		b.nextFree = freeList;
		freeList = b;
	}

	PACKAGE void receive(DatagramSocket socket)
		throws IOException
	{
#ifndef BENCH
		udpPacket.setLength(data.length);
		socket.receive(udpPacket);
		len = udpPacket.getLength();
#else //BENCH
		InChannel.cheatLock.block();
		len = InChannel.cheatLen;
		System.arraycopy(InChannel.cheatBuffer, 0, data, 0, len);
#endif //BENCH
	}

	PACKAGE int length()
	{
		return len;
	}
}
