/* 
 * RouteTable.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ROUTE_PKG;

import JANOS_CORE_PKG.Address;
import JANOS_CORE_PKG.AddressFormatException;
import JANOS_CORE_PKG.NoSuchNodeAddress;

import java.io.FileInputStream;
import java.io.StreamTokenizer;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * Routing table.  Maps 32-bit integer "active addresses" to
 * RouteEntries.  A RouteEntry contains the core Address object
 * that is the "real" address of the target node.
 * <p>
 * RouteEntries are tracked as neighbors and for each known destination.
 *
 * @author David Wetherall
 */
public class RouteTable 
	extends Hashtable
{
	// some RIP-like constants
	public final static int DELETEINTERVAL = 180000;
	public final static int GCINTERVAL = 120000;
	public final static int UPDATEINTERVAL = 30000;
	public final static int UPDATESPREAD = 5000;
	public final static int TRIGGERINTERVAL = 3000;
	public final static int TRIGGERSPREAD = 2000;
	public final static int MAXROUTESPERUPDATE = 64;
	public final static int INFINITY = 16;
	
	public final static boolean LOGGING = false;
	
	/**
	 * Logical address of the node this table is on.
	 */
	public final int node;
	
	/**
	 * A table of all immediate neighbors and the
	 * next hop to get to that neighbor.
	 */
	private final Hashtable neighbor_table; //ulana
	
	// store of timer events
	private final Vector events;
	
	// gc pool, unavailable for forwarding
	private final Hashtable gcpool;
	
	// gateways: hardwired connections to other routes
	private final Hashtable gws;
	
	// degrade route events clock to one tick per second
	private final static long RCRESOLUTION = 1000;
	
	// whether dynamic routing
	private final boolean updates;
	
	// used for suppressing triggered updates
	private boolean trigger;
	
	public RouteTable(int addr, boolean doUpdates) {
		super();
		node = addr;
		updates = doUpdates;
		neighbor_table = new Hashtable(); //ulana
		events = new Vector();
		gcpool = new Hashtable();
		gws = new Hashtable();
	}
	
	/**
	 * Return the entire table of known nodes
	 *
	 * @return Enumeration set of neighbors, each is a RouteEntry
	 *
	 * @see RouteEntry
	 */
	public Enumeration getNeighbors()
	{
		return neighbor_table.elements(); 
	}

	public int neighborCount()
	{
		return neighbor_table.size(); 
	}
	
	public boolean getUpdates()
	{
		return updates; 
	}
	
	public Vector getEvents()
	{
		return events; 
	}
	
	public boolean getTrigger()
	{
		return trigger;
	}
	
	public Hashtable getGCPool()
	{
		return gcpool; 
	}
	
	public Hashtable getGWs()
	{
		return gws;
	}
	
	// compute a route event clock time
	private long routeClock(long base, long add, long spread) {
		long time = base + add + (long)((double)spread*(Math.random()-0.5));
		time += RCRESOLUTION - 1 ;
		time -= (time % RCRESOLUTION);
		return time;
	}
	
	private RouteEntry castToRouteEntry(Object o) {
		if (o instanceof RouteEntry) {
			return (RouteEntry)o;
		}
		return null;
	}
	
	public RouteEntry put(int key, RouteEntry element) {
		int nexthop = element.getLogicalNext();
		
		// If the route describes a route to an immediate neighbor
		// (dest == nexthop) add to the neighbor table.
		if (key == nexthop)
			neighbor_table.put(new Integer(nexthop), element);
		
		// Add to the real table of host -> route mapping
		return castToRouteEntry(super.put(new Integer(key), element));
	}
	
	public RouteEntry get(int key) {
		return castToRouteEntry(super.get(new Integer(key)));
	}
	
	public RouteEntry remove(int key) {
		return castToRouteEntry(super.remove(new Integer(key)));
	}
	
	public void read(String file) 
		throws IOException, AddressFormatException, NoSuchNodeAddress
	{
		FileInputStream f = new FileInputStream(file); 
		StreamTokenizer st =
			new StreamTokenizer(new BufferedReader(new InputStreamReader(f)));

		st.resetSyntax();
		st.whitespaceChars(0, ' ');
		st.wordChars('0', '9');
		st.wordChars('.', '.');
		st.wordChars(':', ':');
		st.wordChars('-', '-');
		st.wordChars('A', 'Z');
		st.wordChars('a', 'z');
		st.commentChar('#');
		
		int src = 0;
		int dst = 0;
		int nxt = 0;
		String addr = null;
		
		TRACE(Trace.ROUTE_READ_TABLE, "Reading route table for node " +Address.toString(node));

		while (st.nextToken() == StreamTokenizer.TT_WORD) {
			if (src == 0) {
				assert(dst == 0);
				assert(nxt == 0);
				assert(addr == null);
				src = Address.fromString(st.sval);
			}
			else if (dst == 0) {
				assert(nxt == 0);
				assert(addr == null);
				dst = Address.fromString(st.sval);
			} 
			else if (nxt == 0) {
				assert(addr == null);
				nxt = Address.fromString(st.sval);
			}
			else {
				assert(addr == null);
				addr = st.sval;
			
				// If this line is for a route from this node
				if (src == node) {
					if (!updates) {
						TRACE(Trace.ROUTE_READ_TABLE, "\tAdding route to " +Address.toString(dst)+ " via " +Address.toString(nxt)+ " (really " +addr+ ").");
						// if static ignoring cost and everything else
						put(dst, new RouteEntry(dst, nxt, addr));
					} else {
						// if dynamic initialize only direct connects and pin them
						if (dst == nxt) {
							TRACE(Trace.ROUTE_READ_TABLE, "\tAdding direct connection to " +Address.toString(dst));
							
							RouteEntry r = new RouteEntry(dst, nxt, addr);
							gws.put(new Integer(dst), r);
						}	    
					}
				}
				src = 0;
				dst = 0;
				nxt = 0;
				addr = null;
			}
		}
		
		if (src != 0) 
			throw new IOException("premature end of input"); // xxx wrong exeception
		else if (st.ttype != StreamTokenizer.TT_EOF)
			throw new IOException("can't handle " + st.toString());// xxx wrong exeception
	}

#if 0
	
	public void sendUpdate(RouteEntry g, ChannelAddress addr, Vector r) {
		RouteUpdateCapsule u = 
			new RouteUpdateCapsule(server.getAddress(), g.dst, addr, r);
		
		// NOTE: we synchronized on the node long ago
		
		// we bypass the node and send over channel directly
		// since we may be bootstrapping the routing database
		server.context = u;
		server.link.send(u, g.addr);
	}
	
	public RouteEvent findTimer(long when) {
		for (Enumeration e = events.elements(); e.hasMoreElements();) {
			RouteEvent r = (RouteEvent)e.nextElement();
			if (r.when == when) { return r; }
		}
		return null;
	}
	
	public RouteEvent findFirstTimer() {
		if (events.isEmpty()) return null;
		
		RouteEvent low = (RouteEvent)events.firstElement();
		for (Enumeration e = events.elements(); e.hasMoreElements();) {
			RouteEvent r = (RouteEvent)e.nextElement();
			if (r.when < low.when) low = r;
		}
		
		return low;
	}
	
	public RouteEvent setTimer(long when, int type) {
		RouteEvent re = findTimer(when);
		
		if (re == null) {
			re = new RouteEvent(when, type);
			events.addElement(re);
		} else {
			re.what |= type;
		}
		
		if (LOGGING) Entity.info("RouteTable: Adding timer, "+
					 events.size() + " events pending");
		
		return re;
	}
	
	public void clearTimer(RouteEvent r) {
		events.removeElement(r);
		
		if (LOGGING) Entity.info("RouteTable: Removing timer, "+
					 events.size() + " events remain");
	}
	
	public void setDeleteTimer(RouteEntry r, long now) {
		long when = routeClock(now, DELETEINTERVAL, 0);
		r.expires = when;
		RouteEvent re = setTimer(when, RouteEvent.DELETE);
		re.deleteCount++;
	}
	
	public void clearDeleteTimer(RouteEntry r) {
		RouteEvent re = findTimer(r.expires);
		if ((re==null) || ((re.what&RouteEvent.DELETE)==0) || (re.deleteCount==0))
			Entity.error("Bad reference count implementation for DELETE");
		
		if (--re.deleteCount == 0) re.what = re.what & (~RouteEvent.DELETE);
		if (re.what == RouteEvent.EMPTY) clearTimer(re);
	}
	
	public void setGCTimer(RouteEntry r, long now) {
		long when = routeClock(now, GCINTERVAL, 0);
		r.expires = when;
		RouteEvent re = setTimer(when, RouteEvent.GC);
		re.gcCount++;
	}
	
	public void clearGCTimer(RouteEntry r) {
		RouteEvent re = findTimer(r.expires);
		if ((re == null) || ((re.what & RouteEvent.GC)==0) || (re.gcCount == 0))
			Entity.error("Bad reference count implementation for GC");
		
		if (--re.gcCount == 0) re.what = re.what & (~RouteEvent.GC);
		if (re.what == RouteEvent.EMPTY) clearTimer(re);
	}
	
	public void setPeriodicTimer(long now) {
		long when = routeClock(now, UPDATEINTERVAL, UPDATESPREAD);
		setTimer(when, RouteEvent.PERIODIC);
	}
	
	public void setTriggerTimer(long now) {
		long when = routeClock(now, TRIGGERINTERVAL, TRIGGERSPREAD);
		setTimer(when, RouteEvent.TRIGGER);
		trigger = true;
	}
	
	public boolean updateEntry(int dest,int dist,int gw,String link,long now) {
		RouteEntry r = get(dest);
		int cost = dist + 1;
		if (cost > INFINITY) cost = INFINITY;
		
		if (r == null) {
			if (cost != INFINITY) {
				
				try {
					ChannelAddress addr = ChannelAddress.New(link);
					r = new RouteEntry(dest, gw, addr);
					r.cost = cost;
					r.change =true;
					
					// might have to resusitate from gc pool
					RouteEntry gc = (RouteEntry)getGCPool().get(new Integer(dest));
					if (gc != null) {
						clearGCTimer(gc);
						getGCPool().remove(new Integer(dest));
					}
					
					Entity.info("RouteTable: Adding route to "+
						    Address.toString(dest) +" (gw="+
						    Address.toString(gw) + ", cost="+ cost +")");
					
					put(dest, r);
					setDeleteTimer(r, now);
					
				} catch (Exception e) {
					Entity.info("RouteTable: Can't create address "+ link);
				}
				
				return true;
			}
		} else { 
			
			// route exists, see if anything changed as per input processing spec
			if (((r.nxt == gw) && (cost != r.cost)) || (cost < r.cost)) {
				r.nxt = gw;
				r.cost = cost;
				r.change = true;
				
				if (cost != INFINITY) {
					
					// finish updating an existing route
					Entity.info("RouteTable: Updating route to "+
						    Address.toString(dest) +" (gw="+
						    Address.toString(gw) + ", cost="+ cost +")");
					
					clearDeleteTimer(r);
					setDeleteTimer(r, now);
					
				} else {
					
					// finish deleting an existing route
					Entity.info("RouteTable: Removing route to "+
						    Address.toString(dest) +" (gw="+
						    Address.toString(gw) + ")");
					
					clearDeleteTimer(r);
					gcpool.put(new Integer(r.dst), r);
					remove(r.dst);
					r.change = true;
					setGCTimer(r, r.expires);
				}
				
				return true;
			}      
			
			// finally, just reset the timers if confirming update
			if ((r.nxt == gw) && (cost == r.cost)) {
				clearDeleteTimer(r);
				setDeleteTimer(r, now);
			}
		}
		
		// if we reach here, there were no changes
		return false;
	}
	
	public void deleteEntries(RouteEvent e) {
		boolean change;
		
		do {
			change = false;
			for (Enumeration f = this.elements(); f.hasMoreElements();) {
				RouteEntry r = (RouteEntry)f.nextElement();
				if (r.expires == e.when) {
					
					Entity.info("RouteTable: Removing route to "+
						    Address.toString(r.dst) +" (gw="+
						    Address.toString(r.nxt) + ")");
					
					r.cost = INFINITY;
					gcpool.put(new Integer(r.dst), r);
					remove(r.dst);
					r.change = true;
					if (!trigger) setTriggerTimer(r.expires);
					setGCTimer(r, r.expires);
					change = true;
					break;
				}
			}
		} while (change);
	}
	
	public void gcEntries(RouteEvent e) {
		boolean change;
		
		do {
			change = false;
			for (Enumeration f = gcpool.elements(); f.hasMoreElements();) {
				RouteEntry r = (RouteEntry)f.nextElement();
				if (r.expires == e.when) {
					
					if (LOGGING) Entity.info("RouteTable: GCing route to "+
								 Address.toString(r.dst));
					
					gcpool.remove(new Integer(r.dst));
					change = true;
					break;
				}
			}
		} while (change);
	}
	
	public void clearChangeFlags() {
		for (Enumeration f = this.elements(); f.hasMoreElements();) {
			RouteEntry d = (RouteEntry)f.nextElement();
			d.change = false;
		}
		
		for (Enumeration f = getGCPool().elements(); f.hasMoreElements();) {
			RouteEntry d = (RouteEntry)f.nextElement();
			d.change = false;
		}
	}
	
	public void sendRouteWorker(RouteEntry r, ChannelAddress addr, int cause) {
		Vector v = new Vector();
		boolean update = false;
		boolean gcswitch = false;
		int routes = 0;
		int poison = 0;
		boolean gwchange = false;
		
		for (Enumeration f = this.elements(); f.hasMoreElements();) {
			RouteEntry d = (RouteEntry)f.nextElement();
			
			// we search the gc pool after the main table
			if (!gcswitch & !f.hasMoreElements()) {
				f = getGCPool().elements();
				gcswitch = true;
			}
			
			// skip the route to that gateway
			if (d.dst == r.dst) {
				if (d.change) gwchange = true;
				continue;
			}
			
			// only send changes for trigger
			if ((cause == RouteEvent.TRIGGER) && !d.change) continue;
			
			// check for split horizon with poison
			int cost = d.cost;
			if (d.nxt == r.dst) {
				cost = INFINITY;
				poison++;
			}
			
			RouteEntry u = new RouteEntry(d.dst, d.nxt, d.addr);
			u.cost = cost;
			v.addElement(u);
			routes++;
			
			// sent off a full message
			if (routes == MAXROUTESPERUPDATE) {
				
				if (LOGGING) Entity.info("RouteTable: Sending "+ routes + 
							 " routes ("+ poison + " poison) to "+
							 Address.toString(r.dst));
				
				sendUpdate(r, addr, v);
				v.removeAllElements();
				routes = 0;
				poison = 0;
				update = true;
			}
		}
		
		// and send the final batch, which may be empty for periodic update
		// or a trigger update if the route to the gateway changed
		if (routes > 0 || 
		    (!update && (cause == RouteEvent.PERIODIC)) ||
		    (gwchange && (cause == RouteEvent.TRIGGER))) {
			
			if (LOGGING) Entity.info("RouteTable: Sending "+ routes + 
						 " routes ("+ poison + " poison) to "+
						 Address.toString(r.dst));
			
			sendUpdate(r, addr, v);
			v.removeAllElements();
			routes = 0;
			poison = 0;
		}
	}
	
	public void sendRouteMessages(int cause) {
		
		// generate messages for each directly connected gateway
		for (Enumeration e = gws.elements(); e.hasMoreElements();) {
			RouteEntry r = (RouteEntry)e.nextElement();      
			sendRouteWorker(r, server.link.source, cause);
		}
		
		// and for additional learned gateways
		for (Enumeration e = this.elements(); e.hasMoreElements();) {
			RouteEntry r = (RouteEntry)e.nextElement();      
			if ((r.cost == 1) && (gws.get(new Integer(r.dst)) == null))
				sendRouteWorker(r, server.link.source, cause);
		}
	}
	
	public void run() {
		if (!updates)
			return;
		
		if (LOGGING)
			Entity.info("RouteTable: Firing up timers");
		
		// all processing reached from here and the RouteUpdateCapsule
		// is synchronized on the node itself to avoid concurrent processing;
		// we expect to make progress by releasing this lock during waits
		
		synchronized (server)
		{
			// kick it off after one periodic update
			// to allow everyone to startup
			
			setPeriodicTimer(System.currentTimeMillis());
			
			while (!events.isEmpty()) {
				
				// search the queue to find the nearest event
				RouteEvent next = findFirstTimer();
				
				// if it's time, then process
				long delay = next.when - System.currentTimeMillis();
				if (delay <= 0) {
					
					// sanity check for leftovers in the queue
					if (next.what == RouteEvent.EMPTY) Entity.error("Empty RouteEvent");
					
					if ((next.what & RouteEvent.GC) != 0) {
						if (LOGGING) Entity.info("RouteTable: GC Event");
						
						gcEntries(next);
					}
					
					if ((next.what & RouteEvent.DELETE) != 0) {
						if (LOGGING) Entity.info("RouteTable: DELETE Event");
						
						deleteEntries(next);
					}
					
					if ((next.what & RouteEvent.PERIODIC) != 0) {
						if (LOGGING) Entity.info("RouteTable: PERIODIC Event");
						
						sendRouteMessages(RouteEvent.PERIODIC);
						setPeriodicTimer(next.when);
						
					} else if ((next.what & RouteEvent.TRIGGER) != 0) {
						if (LOGGING) Entity.info("RouteTable: TRIGGER Event");
						
						sendRouteMessages(RouteEvent.TRIGGER);
						clearChangeFlags();
					}
					
					if ((next.what & RouteEvent.TRIGGER) != 0) trigger = false;
					clearTimer(next);
					
				} else {
					
					// otherwise, wait for it to occur or other triggers
					try {
						
						if (LOGGING) Entity.info("RouteTable: Waiting for "+
									 delay + " ms ...");
						server.wait(delay);
					} catch (InterruptedException e) {
						;
					}
				}
			}
			
			// we should always send at least periodic updates
			panic("RouteTable has stopped running");
		}
	}
	
	public void start()
		throws Exception
	{
		new Thread(this).start();
	}
#endif /* 0 */

}
