/* 
 *  $Id: PingAppNoAWT.java,v 1.3 1999/09/07 23:17:15 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *
 * Based on the Ping example developed by David Wetherall for the
 * original ANTS.
 */
// package NULL

import utah.janos.ants.ants.Application;
import utah.janos.ants.ants.Capsule;
import utah.janos.ants.ants.Protocol;
import utah.janos.ants.ants.Node;
import utah.janos.ants.ants.NodeAddress;
import utah.janos.ants.ants.Xdr;
import utah.janos.ants.ants.ByteArray;

import utah.janos.ants.utils.KeyArgs;
import utah.janos.ants.utils.InvalidKeyArgsException;

/**
 * Version of djw's PingApplication w/o GUI
 * <p>
 * Takes most parameters via command line arguments.
 * Accepts the following parameters (in addition to the
 * basic ones accepted by every Application).
 * <ul>
 *  <li> -target <addr>: Address for ping capsules to go to.
 *  <li> -iter <int>: Numer of ping capsules to send [10].
 *  <li> -interv <int>: Inter-capsule delay in ms [5].
 *  <li> -nbuf <int>: Number of packet buffers to provide to the flow [10].
 *  <li> -replenish <boolean>: Not used [false].
 * </ul>
 *
 * @see PingApplication
 * @author Godmar Back
 */

public class PingAppNoAWT
	extends Application
	implements Runnable
{
	private final static String[] defaults = {};
  
	/**
	 * Not used.
	 */
	private boolean replenish = false;
	
	private int target;
	private int rTotCount = 0;
	private int delay;
	private long beginTime;
	
	private int iter   = 4000; // 220 // must be > lossage.
	private int lossage = 20; // 19
	private int interv = 5;
	private int nbuf   = 10;
	
	/**
	 * Invoked when packets are received.
	 */
	synchronized public void receive(Capsule cap)
	{
		long now;
		super.receive(cap);

		now = thisNode().time();

		log(L[6], " ping capsule delivered to App! "
			+ ((PingCapsule) cap).n);

		switch ((rTotCount % 100)) {
		case 0:
			beginTime = now;
			break;
		case 99:
			double diff = 
				(((double)now - (double)beginTime) / 1000.0);
			double thru = 100.0 / diff;
			System.out.println("Thruput= " + thru + " cap/sec");
			
			if (cap instanceof PingCapsule) {
				PingCapsule pcap = (PingCapsule)cap;
				Xdr buf = new Xdr(pcap.getData(), 0);
				long lat = now - buf.LONG();
				System.out.println("Latency= " + lat + " ms");
			}
			System.out.println("Received " + (rTotCount+1) + " packets");
			break;
		}
		
		if (rTotCount >= iter - lossage) {
			System.out.println("Received almost all capsules.  Exiting.");
			System.exit(0); // XXX not safe...  :)
		}
		
		// XXX timeout??
		
		rTotCount++;
	}
	
	public void run()
	{
		System.out.println("sending " 
				   +iter+ " capsules at " 
				   +interv+ "ms intervals");
		
		try
		{
			// First, send a warmup capsule to get the protocols loaded at
			// each node.

			sendCapsule(-1);
//			thisNode().sleep(100);

			for (int i=0; i < iter; i++) {
				sendCapsule(i);
				thisNode().sleep(interv);
			}
		}
		catch (InterruptedException _)
		{
			// eh, whatever.
			System.out.println("interrupted.  Try again.");
		}
	}
	
	private void sendCapsule(int i)
	{
		ByteArray buf = new ByteArray(Xdr.LONG);
		Xdr xdr = new Xdr(buf, 0);
		xdr.PUT(thisNode().time());
		PingCapsule c = new PingCapsule(port, port, target, buf);
		c.n = i;
		send(c);
	}
	
	/**
	 * Parse command line arguments.
	 */
	public void setArgs(KeyArgs k) 
		throws InvalidKeyArgsException
	{
		k.merge(defaults);
		
		try
		{
		for (int i = 0; i < k.length(); i++) {
			if (k.key(i).equals("-target")) {
				target = NodeAddress.fromString(k.arg(i));
				k.strike(i);
			}
			if (k.key(i).equals("-iter")) {
				iter = Integer.parseInt(k.arg(i));
				k.strike(i);
			}
			if (k.key(i).equals("-interv")) {
				interv = Integer.parseInt(k.arg(i));
				k.strike(i);
			}
			if (k.key(i).equals("-nbuf")) {
				nbuf = Integer.parseInt(k.arg(i));
				k.strike(i);
			}
			if (k.key(i).equals("-replenish")) {
				replenish = true; // Boolean.parseBool(k.arg(i));
				k.strike(i);
			}
		}
		}
		catch (Exception e)
		{
			throw new InvalidKeyArgsException("HACK HACK HACK: " +e.getMessage());
		}
		super.setArgs(k);
	}

	/**
	 * Start the Non-GUI ping app.
	 * <p>
	 * Notice how it's different from PingApplication's start.
	 * <p>
	 * WTF sort of comment is that?
	 */
	public void start() 
		throws Exception 
	{
// 		   InChannel pingChannel = 
// 			   new InChannel(thisNode(), new PingProtocol(), "PingProtocol", true);
// 		   
// 		   FlowHandle pingFlow = pingChannel.getFlowHandle();
// 		   
// 		   // provide initial buffer
// 		   pingFlow.addBuffers(nbuf, 8192);
// 		   
// 		   // this will start a new thread - note that this thread is started
// 		   // in this application's context, meaning that it will share CPU time
// 		   // with it.
// 		   pingChannel.start();
// 		   
// 		   new Thread(this).start();
		
		thisNode().register(new PingProtocol());
		thisNode().threadStart(this);
	}
	
	/**
	 * Construct a non-GUI ping app.
	 * <p>
	 * Doesn't do anything.
	 */
	public PingAppNoAWT() 
		throws Exception 
	{
	}
}
