#!/usr/local/bin/perl -w
#
# runtest.pl
#
# Copyright (c) 1999 University of Utah and the Flux Group.
# All rights reserved.
#
# This file is part of Janos.
# 
# Janos is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License, version 2, as
# published by the Free Software Foundation.
# 
# Janos is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, in the file
# COPYING.  If that file is not present, write to the Free
# Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
# 
# The University of Utah and the Flux Group requests users to
# return to flux-dist@cs.utah.edu any improvements that they make.
# 
# Written by Patrick Tullmann <tullmann@cs.utah.edu>
#

#
# Run a regression test.  Output is either diff'd against an output
# file, or a file to diff against is generated.
#
# Usage:
#   runtest.pl [args] <command> <outputfile>
#
# Must use at least one of:
#	--internal-test 
#	--setup
#	--compare
# Optional args:
#	--timeout=N 
#       --logfile=file
#

$diff="/usr/bin/diff -u";

$debuglvl = 0;
$genTest = 0;
$diffTest = 0;

$command = '';
$timeout = 300;

#%diffOpts = {};

$refFile = '';
$logFile = 'log.$$';
$tmpFile = "/tmp/kore-test.$$";
$diffOptStr = '';

while ($_ = $ARGV[0]) {
    last if /^[^-][^-]/; # stop on first arg that doesn't begin with --
    
    shift @ARGV;

    (&internalTest(), exit 0) if /^--internal-test/;
    ($diffTest = 1, next) if /^--compare/;
    ($genTest = 1, next) if /^--setup/;

    if (/^--timeout/) {
	($timeout) = /--timeout=([0-9]*)/;
	next;
    }
    if (/^--debug/) {
	$debuglvl = 1;
	($debuglvl) = /--debug=([0-9]*)/;
	next;
    }
    if (/^--logfile/) {
	($logFile) = /--logfile=([^ ]*)/;
	next;
    }

    if (/^--diffopt/) {
	($diffOptStr) = /--diffopt=([^ ]*)/;
	foreach $opt (split(',', $diffOptStr)) {
	    $diffOpts{$opt} = 1;
	}
	next;
    }

    &usage("Unknown option: $_");
}    

&dbgPrint(1, "timeout: $timeout; logfile: $logFile; refFile: $refFile; tmpFile: $tmpFile; diffOpts: $diffOptStr\n");
    
($diffTest && $genTest) && &usage("Only one of --compare and --setup makes sense.");

($#ARGV >= 1) || &usage("Need to supply a command and a reference file.");

$command = shift @ARGV;
$refFile = shift @ARGV;

# refFile should exist if we're doing $diffTest
if (($diffTest) && (! -T $refFile)) {
	print ("$command cannot be run no reference output file ($refFile)\n");
	exit 0;
}

## If we're generating a reference output, do it now, and then exit
if ($genTest) {
    $rc = &runCmd($command, $refFile);
    exit $rc;
}

# Otherwise, we're supposed to diff the output
if (($logFile eq '') || ($logFile eq '-') || ($logFile eq 'stdout')) {
    open(LOGFILE, ">&STDOUT");
} else {
    open(LOGFILE, ">>$logFile") || die("Cannot open $logFile for appending");
}

#print ("\'$command\' ... ");
&dbgPrint(2, "Running command ($command) output will be in $tmpFile\n");
$rc = &runCmd($command, $tmpFile);

if ($rc != 0) {
    unlink $tmpFile;
    print STDERR "Command \'$command\' failed in some fashion (rc=$rc).\n";
    exit ($rc);
}

die "Command output unavailable, no $tmpFile\n" if ! -T $tmpFile;

&dbgPrint(2, "Generating diff... ($diff $refFile $tmpFile)\n");
open(DIFF, "$diff $refFile $tmpFile|")
    || die ("Couldn't run \'$diff $refFile $tmpFile\': $!\n");
&dbgPrint(2, "Got diff... ($diff $refFile $tmpFile)\n");

if (eof DIFF) {
    ### SUCCESS
    #print ("\'$command\' succeeded.\n");
    printToLog("--------------------------------------------------\n");
    printToLog("Test succeeded --- \'$command\'.\n");
    printToLog("   [BUG FIXED: There's a previous failure!]\n")
	if (-T $refFile . "failure");
    printToLog("--------------------------------------------------\n");
    printToLog( "\n");
    unlink $tmpFile;
} else {
    ### FAILURE
    print ("\'$command\' FAILED.\n");
    $handledFailure = 0;	# gets set if this failure is the same as previous failure

    $failureFile = $refFile . ".failure";
  FAILURECOMPARE:
    while (-T $failureFile) {
	&dbgPrint(2, "Comparing against previous failure output ($failureFile)\n");

	### Compare against last failure
	### sed magic strips out the diff header lines (which are always different)
	#open(LASTFAILDIFF, "$diff $refFile $tmpFile | $diff $failureFile -|")
	open(LASTFAILDIFF, 
	     "$diff $failureFile $tmpFile |")
	    || die "Couldn't diff $failureFile and $tmpFile: $!\n";
	
	## If eof means no differences, so this is nuttin "new"
	if (eof LASTFAILDIFF) {
	    $handledFailure = 1; # no need to save output 
	    ## Same difference in failure as last failure
	    printToLog( "--------------------------------------------------\n");
	    printToLog( "**  Test FAILED --- \'$command\': same failure as $failureFile\n");
	    printToLog( "--------------------------------------------------\n");
	    printToLog( "\n");
	    close (LASTFAILDIFF);
	    last FAILURECOMPARE;
	} 

	if (!defined $chkfailVer) {
	    $chkfailVer = 1;
	} else {
	    $chkfailVer++;
	}
	
	$failureFile = $refFile . ".failure.$chkfailVer";
	close (LASTFAILDIFF);
    }
    
    ## Failure was "handled" if it was the same as previous failure.
    if (! $handledFailure) {
	printToLog( "--------------------------------------------------\n");
	printToLog( "**  Test FAILED --- \'$command\' ==> diff $refFile -\n");
	printToLog( "      [New kind of Failure!]\n");
	printToLog( "\n");

	# Log the diff, so user can see the problem.
	while (<DIFF>) {
	    printToLog($_); 
	} 

	# Generate a uniq file name for the output of this run, if necess.
	if (-T $failureFile) {
	    $failVer = 1;
	    while (-T $failureFile . "$failVer") { $failVer++; }

	    $failureFile .= "." . $failVer
	}

	printToLog( "\n");
	printToLog( "**  Failure output saved in $failureFile \n");
	printToLog( "--------------------------------------------------\n");
	printToLog( "\n");

	# Mv the output to its permanent resting place
	system ("mv $tmpFile $failureFile 1> /dev/null 2>/dev/null");
    } else {
	# if case mv's the file to a permanent spot.  Nuke it here.
	unlink $tmpFile;
    }

    close(DIFF);
}

## ----------------------------------------------------------------------
## ----------------------------------------------------------------------


## This sub just checks that runtest.pl will (probably)
## work on this system.
sub internalTest {

    ## test open of tmp file
    open(TMP, ">$tmpFile") 
	|| die ("Failed, cannot open $tmpFile: $!\n");
    close(TMP);
    unlink $tmpFile;

    ## Test diff
    $rc = (system ("$diff /dev/null /dev/null") / 256);
    die "Failed, couldn't run $diff: $!" if $rc;
    print "Using $diff\n";

    print "Passed internal tests\n";

    ## No more arg parsing, please...
    exit 0;
}

sub usage {
    print STDERR "\n";
    print STDERR "@_\n";
    print STDERR "runtest.pl [args] <command> <outputfile>\n";
    print STDERR "\t[args] should be one of --internal-test --setup --compare\n";
    print STDERR "\n";

    exit 11;
}

$childDied = 0;

sub childHandler {
    $childDied = 1;
}

sub runCmd {
    local($command, $refFile) = @_;
    local($pid, $rc); # local(SAVEOUT, STDOUT, STDERR, REFFILE, CHILDLINK, PARENTLINK);
    local($deadct);

    # global: $timeout


    $childDied = 0;
    $pid = fork;
    if (!defined $pid) {
	die ("Error forking: $!\n");
    }
    elsif ($pid != 0) {
	&dbgPrint(2, "In parent\n");
	$SIG{'CHLD'} = 'childHandler';

	# parent waits for a short while for child to die.
	# (catches SIGCHLD and set childDied to 1)
	&dbgPrint(2, "Parent blocking for $timeout seconds.\n");
	select(undef, undef, undef, $timeout);
	&dbgPrint(2, "Parent woke\n");

	if ($childDied == 0) {
	    &dbgPrint(2, "Child hung, killing (pid=$pid)\n");
	    $deadct = kill 'TERM', $pid;
	    &dbgPrint(1, "Tried to kill $pid, but cannot\n") if ($deadct == 0);

	    ## Write something into the output file
	    open(REFFILE, ">>$refFile");
	    print REFFILE "\n*** Command timed-out ($timeout seconds, command was \'$command\') ***\n\n";
	    close (REFFILE);

	    # clean the child up
	    $deadct = waitpid($pid, 0);
	    &dbgPrint(1, "Wait failure ($deadct), \$? = $?\n") if (! $deadct);

	    # This is an "okay" way to end the child.
	    $rc = 0;
	} else {
	    &dbgPrint(2, "Parent cleaning up dead child.\n");

	    # Clean child up and use its return value as ours
	    waitpid ($pid, 0);
	    $rc = ($? >> 8);

	    &dbgPrint(2, "Parent cleaned up. (rc=$rc)\n");
	    $rc = 0; # XXX
	}

	return $rc;
    } else {
	&dbgPrint(2, "In child (refFile: $refFile; command: \'$command\')\n");

	## Setup any filters/pipes to process the output:
	$pipes = '';
	$pipes .= "|sort" if (defined $diffOpts{'sort'});
	$pipes .= '|perl -p -e \'s/(0x[0-9A-Fa-f]*)/length($1) . NUMBER/ge\'' if (defined $diffOpts{'nukehex'});
	$pipes .= "|sed -e 's/[0-9]/X/g'" if ((defined $diffOpts{'nukedigit'}) || (defined $diffOpts{'nukedigits'}));

	# child runs the command, redirect stdout and stderr to the file
	open(STDOUT, "$pipes>$refFile") || die "Could not open $refFile: $!\n";
	open(STDERR, ">&STDOUT") || die "Could not dup stderr to stdout: $!\n";

	if (!exec("$command")) { 
	    open(STDOUT, ">&SAVEOUT");
	    &dbgPrint(2, "CHILD: Failed exec of \'$command\': $!\n");
	    exit(42);
	}
    }
}

sub dbgPrint {
    local($lvl) = shift @_;
    print STDOUT @_ if ($lvl <= $debuglvl)
}

sub printToLog {
    print LOGFILE @_;
}

